/*

  QQ.cc - QR3 Version

  Date: July 2003

  Purpose:  Provides a user with the ability to build an SQL
            SELECT query without knowing SQL statements.

            The main purpose of the QR3 integrate version is to produce a flat
            table from related tables to act as fodder for a QR3 Report. 



  Based on code by Marc Hamelin

  Adapted and developed by Michael Nuwer

  Adapted for QR3 by Michael Nuwer and Ronnie MacGregor

  Final integration & split to UI and QQ.cc Query engine by Ronnie MacGregor

*/

class QQ(parentObj) of Paintbox(parentObj) custom
   with (this)
      //onClose = CLASS::onClose                                       
      visible = false                             
      parent.metric := 6 // Pixels                   
      left = 0                                       
      top = 0                                        
      width = 25                                     
      height = 25 
   endwith

   // Get array of Alias names
   this.aAliasList = new Array()
   this.aAliasList = _app.databases.current.getSchema("Databases")
   this.aAliasList.sort()

   // Display confirmations
   this.Silent = false

   // -------------------------------------

   function Initialise()

      // Page 1 DataPath
      this.cAlias = ''
      this.bUseAlias = false
      this.cDataPath = ''


      // Page 2 Tables
      this.cTablesExt = ''
      this.aAllTablesList = new array()
      this.aAvailableTables = new array()
      this.aSelectedTables = new array()


      // Page 3 Fields
      this.aAllFieldsList = new array()
      this.aAvailableFields = new array()
      this.aSelectedFields = new array()
      this.bDistinct = false  
      

      // Page 4 Join
      this.aLinkFields = new array()
      this.aLinkTypes = new Array()


      // Page 5 Condition
      this.aConditions = new array()
      this.bIsNot = false
      this.bIsOr  = false


      // Page 6 Order
      this.aOrderFields = new array()


      // Page 7 Group
      this.aGroupFields = new array()


      // Page 8 Custom
      this.CustomSQLValue = ''


      // The Engine Built SQL Query string
      this.SQLValue = '' 

      
      this.SQLisSaved = true
      this.SQLFilePath = ''
      this.SQLFileName = ''


      // Output Table
      this.OPTableName = ''



   return


   // -------------------------------------

   function Open_SQL_File(cFile)

      // Initialise QQ properties
      this.Initialise()

      // Check file exists
      if not file(cFile)
         cTxt = "SQL Query file does not exist :" + chr(13) + chr(13) ;
              + cFile + chr(13) + chr(13) ;
              + "Unable to reload it into Quick Query."
         msgbox(cTxt,"Open SQL File",16)         
         return false
      endif

      oString = new StringParser()

      f = new file()
      fSize = f.size(cFile)  
      f.open(cFile)
      cMaskStr = f.read(fSize)
      f.close()
       
      cMaskStr = oString.ChrSwap(cMaskStr,chr(10)," ")
      cMaskStr = oString.ChrSwap(cMaskStr,chr(13)," ")
      cMaskStr = oString.ChrSwap(cMaskStr,chr(9)," ")

      // remove all double spaces
      cMaskStr = oString.removeSpace(cMaskStr)

      if left(cMaskStr,11)=='--Database:'
         cMaskStr = substr(cMaskStr,at(" ",cMaskStr))
         // cAlias = 
      endif

      cSQLString = cMaskStr

      // In case a key word is inside a quoted string.
      // This method masks the characters inside quotes.
      cMaskStr = oString.MaskString(cMaskStr)
      cMaskStr = upper(cMaskStr)

      aErrors = new array()
      if " IN " $ cMaskStr
         aErrors.add("IN")
      endif
      if " STARTING WITH " $ cMaskStr
         aErrors.add("STARTING WITH")
      endif
      if " CONTAINING " $ cMaskStr
         aErrors.add("CONTAINING")
      endif
      if " BETWEEN " $ cMaskStr
         aErrors.add("BETWEEN")
      endif
      if " HAVING " $ cMaskStr
         aErrors.add("HAVING")
      endif
      if "||" $ cMaskStr
         aErrors.add("Calulated Field")
      endif
      if " * " $ cMaskStr
         aErrors.add("Wild Card Character")
      endif
      if oString.NumToken(cMaskStr,"SELECT ") > 1
         aErrors.add("Sub Query")
      endif

      if aErrors.size > 0
         msg = "The following are not supported by Quick Query 3" +chr(13)
         msg += "Can not load SQL file."  +chr(13)
         for i = 1 to aErrors.size
            msg += aErrors[i]  +chr(13)
         next
         msgbox( msg )
         return
      endif

      f.open(cFile)
      cSQLvalue = f.read(fSize)
      f.seek(0)
      cStr = f.gets()
      f.close()
         
      if left(cStr,11)=='--Database:'
         cAlias = trim(substr(cStr,12))
         local oDatabase,aSchema
         try
            oDatabase = new database()
            oDatabase.databaseName := cAlias
            oDatabase.active := true
            if oDatabase.driverName == "DBASE" 
               this.cTablesExt = ".dbf"
               cDriverType = "LocalFiles"
            elseif oDatabase.driverName == "PARADOX"
               this.cTablesExt = ".db"
               cDriverType = "LocalFiles"
            else
               this.cTablesExt = ""
               cDriverType = "Database"
            endif

         catch(exception e)
            msgbox('Alias type for '+ this.cAlias +' is not supported.','Warning',48)
            oDatabase.active := false
            release object oDatabase
            return
         endtry

         this.SetALias(cAlias)  // Ronnie
             
      else
         this.bUseAlias = false
         this.cAlias = ""
         this.cTablesExt = ".dbf"
      endif

      aPosKey = new array(5,2)  //   Position, Keyword

      // find key words (single)
      aPosKey[1,1] = oString.atToken(cMaskStr,"SELECT ",1)
      aPosKey[1,2] = "SELECT"

      if oString.atToken(cMaskStr,"DISTINCT ",1) > 0
         aPosKey[1,1] = oString.atToken(cMaskStr,"SELECT DISTINCT ",1)
         aPosKey[1,2] = "SELECT DISTINCT"
      endif

      aPosKey[2,1] = oString.atToken(cMaskStr,"FROM ",1)
      aPosKey[2,2] = "FROM"

      aPosKey[3,1] = oString.atToken(cMaskStr,"ORDER BY ",1)
      aPosKey[3,2] = "ORDER BY"

      aPosKey[4,1] = oString.atToken(cMaskStr,"GROUP BY ",1)
      aPosKey[4,2] = "GROUP BY"

      aPosKey[5,1] = oString.atToken(cMaskStr,"WHERE ",1)
      aPosKey[5,2] = "WHERE"

      if oString.atToken(cMaskStr,"WHERE NOT ",1) > 0
         aPosKey[5,1] = oString.atToken(cMaskStr,"WHERE NOT ",1)
         aPosKey[5,2] = "WHERE NOT"
      endif


      x=6
      nCnt = oString.NumToken(cMaskStr,"ON ")
      for i = 1 to nCnt
         aPosKey.grow(1)
         aPosKey[x,1] = oString.atToken(cMaskStr,"ON ",i)
         aPosKey[x,2] = "ON"
         x++
      next
      nCnt = oString.NumToken(cMaskStr,"INNER JOIN ")
      for i = 1 to nCnt
         aPosKey.grow(1)
         aPosKey[x,1] = oString.atToken(cMaskStr,"INNER JOIN ",i)
         aPosKey[x,2] = "INNER JOIN"
         x++
      next
      nCnt = oString.NumToken(cMaskStr,"LEFT OUTER JOIN ")
      for i = 1 to nCnt
         aPosKey.grow(1)
         aPosKey[x,1] = oString.atToken(cMaskStr,"LEFT OUTER JOIN ",i)
         aPosKey[x,2] = "LEFT OUTER JOIN"
         x++
      next
      nCnt = oString.NumToken(cMaskStr,"AND ")
      for i = 1 to nCnt
         aPosKey.grow(1)
         aPosKey[x,1] = oString.atToken(cMaskStr,"AND ",i)
         aPosKey[x,2] = "AND"
         x++
      next
      nCnt = oString.NumToken(cMaskStr,"OR ")
      for i = 1 to nCnt
         aPosKey.grow(1)
         aPosKey[x,1] = oString.atToken(cMaskStr,"OR ",i)
         aPosKey[x,2] = "OR"
         x++
      next

      local bHasGroup 
      bHasGroup = false
      if aPosKey[4,1] > 0 // if group by exists
         bHasGroup = true

         /*
         nCnt = oString.NumToken(cMaskStr,"SUM")
         for i = 1 to nCnt
            aPosKey.grow(1)
            aPosKey[x,1] = oString.atToken(cMaskStr,"SUM",i)
            aPosKey[x,2] = "SUM"
            x++
         next
         nCnt = oString.NumToken(cMaskStr,"COUNT")
         for i = 1 to nCnt
            aPosKey.grow(1)
            aPosKey[x,1] = oString.atToken(cMaskStr,"COUNT",i)
            aPosKey[x,2] = "COUNT"
            x++
         next
         nCnt = oString.NumToken(cMaskStr,"MIN")
         for i = 1 to nCnt
            aPosKey.grow(1)
            aPosKey[x,1] = oString.atToken(cMaskStr,"MIN",i)
            aPosKey[x,2] = "MIN"
            x++
         next
         nCnt = oString.NumToken(cMaskStr,"MAX")
         for i = 1 to nCnt
            aPosKey.grow(1)
            aPosKey[x,1] = oString.atToken(cMaskStr,"MAX",i)
            aPosKey[x,2] = "MAX"
            x++
         next
         nCnt = oString.NumToken(cMaskStr,"AVE")
         for i = 1 to nCnt
            aPosKey.grow(1)
            aPosKey[x,1] = oString.atToken(cMaskStr,"AVE",i)
            aPosKey[x,2] = "AVE"
            x++
         next
         */
      endif 

      aPosKey.sort( ) // sort by col 1
                      // by position of keyWord
                       
      //for i=1 to aPosKey.size/2
      //   ?aPosKey[i,1]
      //   ?? aPosKey[i,2] at 20
      //next

      // used in Where clause
      this.bIsNot=false
      this.bIsOr =false
      for i=1 to aPosKey.size / 2
         if aPosKey[i,1]>0
            cKeyWord  = aPosKey[i,2]
            cKeyValue = substr(cSQLString, aPosKey[i,1] )
            if i = aPosKey.size /2  // if last key
               cKeyValue = substr(cKeyValue,1)
            else   // all other keys
               cKeyValue = substr(cKeyValue,1,aPosKey[i+1,1]- aPosKey[i,1] - len(aPosKey[i+1,2]-1))
            endif
     //       ? cKeyWord
     //       ?? cKeyValue at 12

            if cKeyWord == "SELECT DISTINCT"  and not bHasGroup
               this.bDistinct := true  
               //  fill the fields array from the SELECT
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               do while at(",",cKeyValue) > 0
                  cVar = left(cKeyValue,at(",",cKeyValue)-1)
                  cKeyValue = substr(cKeyValue,at(",",cKeyValue)+1)
                  this.aSelectedFields.add(ltrim(rtrim(cVar)))
               enddo
               // add last field
               this.aSelectedFields.add(ltrim(rtrim(cKeyValue)))

            elseif cKeyWord == "SELECT"   and not bHasGroup
               //  fill the fields array from the SELECT
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")

               do while at(",",cKeyValue) > 0
                  cVar = left(cKeyValue,at(",",cKeyValue)-1)
                  cKeyValue = substr(cKeyValue,at(",",cKeyValue)+1)
                  this.aSelectedFields.add(ltrim(rtrim(cVar)))
               enddo
               // add last field
               this.aSelectedFields.add(ltrim(rtrim(cKeyValue)))

            elseif cKeyWord == "SELECT"   and bHasGroup
               //  fill the fields array from the SELECT
               aGroup = new array()
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")

               do while at(",",cKeyValue) > 0
                  cVar = left(cKeyValue,at(",",cKeyValue)-1)
                  cKeyValue = substr(cKeyValue,at(",",cKeyValue)+1)
                  aGroup.add(ltrim(rtrim(cVar)))
               enddo
               // add last field
               aGroup.add(ltrim(rtrim(cKeyValue)))

               for j = 1 to aGroup.size
                  if at('(' , aGroup[j]) > 0
                     this.aGroupFields.add(ltrim(rtrim(aGroup[j])))
                     cVar = substr(aGroup[j],at('(',aGroup[j])+1)
                     cVar = substr(cVar,1,at(')',cVar)-1)
                     this.aSelectedFields.add(ltrim(rtrim(cVar)))
                  else
                     this.aSelectedFields.add(ltrim(rtrim(aGroup[j])))
                  endif
               next

            elseif cKeyWord == "GROUP BY"                
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               do while at(",",cKeyValue)>0
                  cVar = substr( cKeyValue, 1, at(",",cKeyValue)-1 )
                  this.aGroupFields.add("GROUP BY( "+ ltrim(trim(cVar)) + " )" )
                  cKeyValue = substr(cKeyValue,at(",",cKeyValue)+1)
               enddo
               this.aGroupFields.add("GROUP BY( " + ltrim(trim(cKeyValue))+ " )" )
         
            elseif cKeyWord == "FROM"

               do while at(",",cKeyValue)>0
                  cVar = subStr(cKeyValue,1, at(",",cKeyValue)-1 )
                  cVar = ltrim(trim(cVar))
                  if at([" ],cVar)> 0
                     cVar = substr( cVar, at('" ',cVar)+2 )
                  else
                     cVar = substr( cVar, at(' ',cVar)+1 )
                  endif

                  cVar = oString.ChrSwap(cVar,'"',"")
                  this.aSelectedTables.add(trim(cVar))
                  cKeyValue = substr(cKeyValue,at(",",cKeyValue)+1)
               enddo
               cVar = ltrim(trim(cKeyValue))
               if at([" ],cVar)> 0
                  cVar = substr( cVar, at('" ',cVar)+2 )
               else
                  cVar = substr( cVar, at(' ',cVar)+1 )
               endif
               cVar = oString.ChrSwap(cVar,'"',"")
               this.aSelectedTables.add(ltrim(trim(cVar)))

               this.aAvailableTables = new array()  //////////////////////////////////////////////// redundant ? 
               if not this.bUseAlias
                  cKeyValue = ltrim(cKeyValue)
                  cVar = substr(cKeyValue,2,rat('\',cKeyValue)-1)
                  this.cDataPath = trim(cVar)
                  //"C:\dB2KTutorial\Final\tables\Customer.dbf" Customer,
                  this.cTablesExt = substr(cKeyValue,at(".",cKeyValue))
                  this.cTablesExt = substr(this.cTablesExt,1,at(["],this.cTablesExt)-1)
                  this.DirTables = new array() ////////////////////////////////////////////// redundant ?
                  this.DirTables.dir(this.cDataPath + "*" + this.cTablesExt)
                  for j = 1 to (this.DirTables.size / 5)
                     //this.aAvailableTables.add(proper(left(this.DirTables[j,1],len(this.DirTables[j,1]) - len(this.cTablesExt) -1 )))
                     this.aAvailableTables.add(proper(left(this.DirTables[j,1],at(".",this.DirTables[j,1]) - 1 )))
                  next
               else
                  aSchema = oDatabase.getSchema( "Tables" )
                  for j = 1 to aSchema.size
                     if rat(".",aSchema[j])>0
                        nLen = rat(".",aSchema[j])-1
                        this.aAvailableTables.add(proper(left(aSchema[j],nLen)))
                     else
                        this.aAvailableTables.add(proper(aSchema[j]))
                     endif
                  next
               endif

            elseif cKeyWord == "INNER JOIN"
               cVar = ltrim(trim(cKeyValue))
               if at([" ],cVar)> 0
                  cVar = substr( cVar, at('" ',cVar)+2 )
               else
                  cVar = substr( cVar, at(' ',cVar)+1 )
               endif
               cVar = oString.ChrSwap(cVar,'"',"")
               this.aSelectedTables.add(trim(cVar))
               this.aLinkTypes.add("INNER JOIN")

            elseif cKeyWord == "LEFT OUTER JOIN"
               cVar = ltrim(trim(cKeyValue))
               if at([" ],cVar)> 0
                  cVar = substr( cVar, at('" ',cVar)+2 )
               else
                  cVar = substr( cVar, at(' ',cVar)+1 )
               endif
               cVar = oString.ChrSwap(cVar,'"',"")
               this.aSelectedTables.add(trim(cVar))
               this.aLinkTypes.add("LEFT OUTER JOIN")

            elseif cKeyWord == "ON"
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               cVar = substr( cKeyValue, 2, at(')',cKeyValue) - 2 )
               this.aLinkFields.add(trim(cVar))
               //(Customer."Customer ID" = Invoice."Customer ID")
        
            elseif cKeyWord == "WHERE"
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               if at('(',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at('(',cKeyValue),1,"")
               endif
               if at(')',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at(')',cKeyValue),1,"")
               endif
               this.aConditions.add(ltrim(trim(cKeyValue)))

            elseif cKeyWord == "WHERE NOT"
               this.bIsNot = true
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               if at('(',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at('(',cKeyValue),1,"")
               endif
               if at(')',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at(')',cKeyValue),1,"")
               endif
               this.aConditions.add(ltrim(trim(cKeyValue)))
               this.bIsNot := true ////////////////////////////////////////// Ronnie

            elseif cKeyWord == "AND"
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               if at('(',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at('(',cKeyValue),1,"")
               endif
               if at(')',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at(')',cKeyValue),1,"")
               endif
               this.aConditions.add(ltrim(trim(cKeyValue)))

            elseif cKeyWord == "OR"
               this.bIsOr = true
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               if at('(',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at('(',cKeyValue),1,"")
               endif
               if at(')',cKeyValue) > 0
                  cKeyValue = stuff( cKeyValue, at(')',cKeyValue),1,"")
               endif
               this.aConditions.add(cKeyValue)
               this.bIsOr = true ////////////////////////////////////////// Ronnie
                
            elseif cKeyWord == "ORDER BY"
               // Order by clause
               cKeyValue = oString.ChrSwap(cKeyValue,["],"")
               do while at(",",cKeyValue)>0
                  cVar = substr( cKeyValue, 1, at(",",cKeyValue)-1 )
                  this.aOrderFields.add(ltrim(trim(cVar)))
                  cKeyValue = substr(cKeyValue,at(",",cKeyValue)+1)
               enddo
               this.aOrderFields.add(ltrim(trim(cKeyValue)))
            endif

         endif
      next

      //if this.bUseAlias
         //oDatabase.active := false ////////////////////////////////////////////////////
         //release object oDatabase
      //endif

      for i = 1 to this.aSelectedTables.size
         seekResult = this.aAvailableTables.scan(this.aSelectedTables[i])
         if seekResult > 0
            this.aAvailableTables.delete(seekResult)
            this.aAvailableTables.size -= 1
         endif
      next

      this.aAvailableTables.sort()


      this.aAllFieldsList = new array() //////////////////////////////////redundant ?

      this.SQLisSaved = true

      this.SQLFilePath := substr(cFile, 1, rat("\", cFile, 1) - 1)
      this.SQLFileName := substr(cFile, rat("\", cFile, 1) + 1)

      // ?  " this.SQLFilePath := *" + this.SQLFilePath + "*"  ////////////////////////////////////
      // ?  " this.SQLFileName := *" + this.SQLFileName + "*" /////////////

      this.BuildFieldsList()

   return true

   // -------------------------------------

   function BuildFieldsList()

      // Initialise array
      this.aAllFieldsList = new array()

      // Prepare query
      local q, d
      q = new query()
      if this.bUseAlias
         d = new database()
         d.databaseName = this.cAlias
         d.active = true
         q.database = d
         cTablesPath = ""
      else
         cTablesPath = this.cDataPath + "\"
      endif

      // Add fields from all selected tables to the All Fields array
      for nCount = 1 to this.aSelectedTables.size
         q.sql = "SELECT * FROM '" + cTablesPath + this.aSelectedTables[nCount] + this.cTablesExt + "'"
         q.active = true
         for nCount2 = 1 to q.rowset.fields.size
            if q.rowset.fields[nCount2].baseClassName # "LOCKFIELD"
               cField = q.rowset.fields[nCount2].fieldName
               this.aAllFieldsList.add(this.aSelectedTables[nCount] + "." + cField)
            endif
         next
         q.active = false
      next

      // Update Available Fields array
      for nCount = 1 to this.aAllFieldsList.size
         nTableNo = this.aSelectedTables.scan(substr(this.aAllFieldsList[nCount], 1, at(".", this.aAllFieldsList[nCount])-1))
         if substr(this.aAllFieldsList[nCount], 1, at(".", this.aAllFieldsList[nCount])-1) == this.aSelectedTables[nTableNo]
            cField = substr(this.aAllFieldsList[nCount], at(".",this.aAllFieldsList[nCount]) + 1)
            if this.aSelectedFields.scan(this.aAllFieldsList[nCount]) == 0 
               this.aAvailableFields.add(this.aSelectedTables[nTableNo] + "." + cField)
            endif
         endif
      next 

   return

   // -------------------------------------

   function SetAlias(cAlias)

      // Deal with nothing to process
      if not Len(cAlias) > 0
         msgbox('No Alias name provided','Warning',48)
         return
      endif

      // Check if save required
      if not this.SQLisSaved == true
         cText = "The query has not been saved. " + chr(13) ;
               + "Do you want to save it?"
         nAns =  msgbox(cText,"Save Query",51)
         if  nAns == 6 // Yes
            this.SaveQuery()
         elseif nAns == 2 // Cancel
            return false
         endif
      endif

      // Check Alias supported
      local oDatabase,aSchema
      try
         oDatabase = new database()
         oDatabase.databaseName := cAlias
         oDatabase.active := true

         if oDatabase.driverName == "DBASE" 
            cTablesExt = ".dbf"
            cDriverType = "LocalFiles"
         elseif oDatabase.driverName == "PARADOX"
            cTablesExt = ".db"
            cDriverType = "LocalFiles"
         else
            cTablesExt = ""
            cDriverType = "Database"
         endif

         /*
         if oDatabase.driverName <> "DBASE"
            msgbox('Alias type for '+ oDatabase.driverName +' is not supported.','Warning',48)
            oDatabase.active := false
            release object oDatabase
            this.value = this.cAlias
            return
         endif
        */

      catch(exception e)
         msgbox('Alias type for '+ cAlias +' is not supported.','Warning',48)
         oDatabase.active := false
         release object oDatabase
         return
      endtry


      // Check for valid Alias
      if cDriverType == "LocalFiles"
         cAlias = Upper(Trim(cAlias))         
         cPath = this.FindAliasPath(cAlias)
         if empty(cPath)
            msgbox('Alias '+ cAlias +' not found. Please re-enter','Warning',48)
            return
         elseif cPath = "NOTSUPPORTED"
            msgbox('Alias type for '+ cAlias +' is not supported.','Warning',48)
            return
         endif
      endif

      aSchema = oDatabase.getSchema( "Tables" )
      oDatabase.active := false
      release object oDatabase

      // Initialise QQ
      this.Initialise()
      this.cAlias := cAlias
      if cDriverType == "LocalFiles"
         this.cTablesExt := cTablesExt
         this.cDataPath := trim(cPath) 
      endif

      // Update Tables array
      if cDriverType == "LocalFiles"
         aDirTables = new array()
         aDirTables.dir(this.cDataPath + "\*.db*") // this includes dbt
         for nCount = 1 to (aDirTables.size / 5)
            if not upper(right(aDirTables[nCount,1],3)) == "DBT"
               this.aAllTablesList.add(left(aDirTables[nCount,1], at(".", aDirTables[nCount,1]) - 1))
               this.aAvailableTables.add(left(aDirTables[nCount,1], at(".", aDirTables[nCount,1]) - 1))
            endif
         endfor // next
      else
         for nCount = 1 to aSchema.size
            if rat(".",aSchema[nCount])>0
               nLen = rat(".",aSchema[nCount])-1
               this.aAllTablesList.add(left(aSchema[nCount],nLen))
               this.aAvailableTables.add(left(aSchema[nCount],nLen))
            else
               this.aAllTablesList.add(aSchema[nCount])
               this.aAvailableTables.add(aSchema[nCount])
            endif
         next
      endif
      this.aAllTablesList.sort()
      this.aAvailableTables.sort()

      this.bUseAlias = true

      // Update changed flag
      this.SQLisSaved = false

   return

   // -------------------------------------

   function FindAliasPath

      parameters cDatabase

      /*

      This function returns the path for the alias supplied
      as a parameter (null if it does not exist)

      Based on code posted by Romain Strieff 
      in the Visual dBASE Newsgroups

      Modified to accept a string for an alias name
      instead of a database object by Gary White

      */

      local oDatabase
      
      cRet = null

      if type("cDatabase") == "C" and not empty(cDatabase)
         if Class::DoesAliasExist( cDatabase )
            if type("GetAliasPath") # "FP"
               extern cint GetAliasPath(cHandle,cString,;
                  cString,cString);
                  IDAPI32 from "DbiFormFullName"
            endif
            cRet=space(261)

            try
               oDatabase = new database()
               oDatabase.databaseName := cDatabase
               oDatabase.active := true
               if ( GetAliasPath(oDatabase.handle,"DummyTableName","",cRet ) == 0 )
                  cRet=cRet.left(cRet.lefttrim().lastindexof("\"))
               else
                  cRet=""
               endif
               oDatabase.active := false
               release object oDatabase
            catch(exception e)
               cRet = "NOTSUPPORTED"
            endtry
         endif
      endif

      return cRet

   // -------------------------------------

   function DoesAliasExist( cDatabase )

      // Extracted from BDEAlias.cc by Romain Strieff and Gary White
      // Code by Gary White

      local d, dbArray, lFound, i
      d = new database()
      dbArray=d.getschema( "Databases" )
      release object d
      lFound = false
      for i = 1 to dbArray.size
         if upper( cDatabase ) == upper( dbArray[i] )
            lFound := true
            exit
         endif
      endfor

      return lFound

   // -------------------------------------

   function SetDataPath(cPath)

      // Deal with nothing to process
      if not Len(cPath) > 0
         msgbox('No Data Path provided','Warning',48)
         return
      endif

      // Check if save required
      if not this.SQLisSaved == true
         cText = "The query has not been saved. " + chr(13) ;
               + "Do you want to save it?"
         nAns =  msgbox(cText,"Save Query",51)
         if  nAns == 6 // Yes
            this.SaveQuery()
         elseif nAns == 2 // Cancel
            return false
         endif
      endif

      // Set the data directory
      this.Initialise()
      this.cDataPath = cPath 

      // Update Tables array
      aDirTables = new array()
      aDirTables.dir(this.cDataPath + "\*.db*") // this includes dbt
      for nCount = 1 to (aDirTables.size / 5)
         if not upper(right(aDirTables[nCount,1],3)) == "DBT"
            this.aAllTablesList.add(left(aDirTables[nCount,1], at(".", aDirTables[nCount,1]) - 1))
            this.aAvailableTables.add(left(aDirTables[nCount,1], at(".", aDirTables[nCount,1]) - 1))
         endif
      endfor // next
      this.aAvailableTables.sort()

      // Check we have some tables
      if this.aAllTablesList.size > 0
         this.cTablesExt = lower(substr(aDirTables[1,1], at(".", aDirTables[1,1])))
      else
         msgbox("There are no Data Tables in " + cPath, "Warning", 48)
      endif

      // Set changed flag
      this.SQLisSaved = false

   return

   // -------------------------------------

   function Generate_Query()

      local cString, newSQLquery,oStrParse
      oStrParse = new stringParser()

      newSQLQuery = ""

      cTablesPath = this.cDataPath + "\" 
      if this.bUseAlias == true
         newSQLquery += "--Database:" + this.cAlias + chr(13) + chr(10)
         cTablesPath = ""
      endif

      newSQLquery += "SELECT "
      if this.bDistinct == true
         newSQLquery += "DISTINCT "
      endif

      if this.aGroupFields.size = 0
         first = true
         for i = 1 to this.aSelectedFields.size
            cField = oStrParse.addQuotes(this.aSelectedFields[i])
            if first = true
               newSQLquery += cField
               first = false
            else
               newSQLquery += ", " + cField
            endif
         next
      else
         first = true
         for i = 1 to this.aGroupFields.size

            if "GROUP BY" $ this.aGroupFields[i]
               cString = substr(this.aGroupFields[i],at("(",this.aGroupFields[i])+2)
               cString = substr(cString,1,at(")",cString)-2)
            else
               cString = this.aGroupFields[i]
            endif

            cString = oStrParse.addQuotes(cString)

            if first = true
               newSQLquery += cString
               first = false
            else
               newSQLquery += ", " + cString
            endif
         next
      endif
      processLink = new array()
      first = true
      for i = 1 to this.aSelectedTables.size
         found = false
         for cnt = 1 to this.aLinkFields.size
            currentlink = right(this.aLinkFields[cnt],len(this.aLinkFields[cnt]) - at("=",this.aLinkFields[cnt]) - 1)
            curtablelink2 = left(currentlink,at(".",currentlink) - 1)
            if this.aSelectedTables[i] == curtablelink2
               found = true
            endif
         next
         if found = true
            processLink.add(this.aSelectedTables[i])
         else
            if first = true
               if empty(this.cTablesExt) // database server
                  newSQLquery += " " + chr(13)+chr(10) + 'FROM ' + this.aSelectedTables[i] + ' ' + this.aSelectedTables[i]
               else
                  newSQLquery += " " + chr(13)+chr(10) + 'FROM "' + cTablesPath + this.aSelectedTables[i] + this.cTablesExt + '" ' + this.aSelectedTables[i]
               endif
               first = false
            else
               if empty(this.cTablesExt) // database server
                  newSQLquery += ', ' + this.aSelectedTables[i] + ' ' + this.aSelectedTables[i]
               else
                  newSQLquery += ', "' + cTablesPath + this.aSelectedTables[i] + this.cTablesExt + '" ' + this.aSelectedTables[i]
               endif
            endif
         endif
      next
      for i = 1 to processLink.size
         for cnt = 1 to this.aLinkFields.size
            currentlink = right(this.aLinkFields[cnt],len(this.aLinkFields[cnt]) - at("=",this.aLinkFields[cnt]) - 1)
            curtablelink2 = left(currentlink,at(".",currentlink) - 1)
            if processLink[i] == curtablelink2

               if empty(this.cTablesExt) // database server
                  newSQLquery += " " + chr(13)+chr(10) + this.aLinkTypes[cnt] + ' ' + processLink[i] + ' ' + processLink[i]
               else
                  newSQLquery += " " + chr(13)+chr(10) + this.aLinkTypes[cnt] + ' "' + cTablesPath + processLink[i] + this.cTablesExt + '" ' + processLink[i] 
               endif
               cLeft  = oStrParse.addQuotes(substr(this.aLinkFields[cnt],1,at(" =",this.aLinkFields[cnt])-1))
               cRight = oStrParse.addQuotes(substr(this.aLinkFields[cnt],at("= ",this.aLinkFields[cnt])+2))

               newSQLquery += " " + chr(13)+chr(10)+ space(4) + " ON (" + cLeft+" = "+cRight + ")"
            endif
         next
      next
      

      // Deal with Conditions

      first = true
      for i = 1 to this.aConditions.size
         cField = this.aConditions[i]
         cCond = ""

         //  The order of the following tests is VERY important

         if at("<>",cField)>0
            cField = substr(cField,1,at("<>",cField)-2)
            cCond = substr(this.aConditions[i],at(" <>",this.aConditions[i]))
         elseif at(">=",cField)>0
            cField = substr(cField,1,at(">=",cField)-2)
            cCond = substr(this.aConditions[i],at(" >=",this.aConditions[i]))
         elseif at("<=",cField)>0
            cField = substr(cField,1,at("<=",cField)-2)
            cCond = substr(this.aConditions[i],at(" <=",this.aConditions[i]))
         elseif at("<",cField)>0
            cField = substr(cField,1,at("<",cField)-2)
            cCond = substr(this.aConditions[i],at(" <",this.aConditions[i]))
         elseif at("=",cField)>0
            cField = substr(cField,1,at("=",cField)-2)
            cCond = substr(this.aConditions[i],at(" =",this.aConditions[i]))
         elseif at(">",cField)>0
            cField = substr(cField,1,at(">",cField)-2)
            cCond = substr(this.aConditions[i],at(" >",this.aConditions[i]))
         elseif at(" LIKE ",cField)>0
            cField = substr(cField,1,at(" LIKE ",cField)-1)
            cCond = substr(this.aConditions[i],at(" LIKE ",this.aConditions[i]))
         elseif at(" IS NULL",cField)>0
            cField = substr(cField,1,at(" IS NULL",cField)-1)
            cCond = substr(this.aConditions[i],at(" IS NULL",this.aConditions[i]))
         elseif at(" IS NOT NULL",cField)>0
            cField = substr(cField,1,at(" IS NOT NULL",cField)-1)
            cCond = substr(this.aConditions[i],at(" IS NOT NULL",this.aConditions[i]))
         endif
         cField = oStrParse.addQuotes(cField)
         cField = cField + cCond
         if first = true
            if this.bIsNot == true  /////////////////////////////////////////////////// Ronnie
               newSQLquery += " " + chr(13)+chr(10) + "WHERE NOT ( " + cField 
            else
               newSQLquery += " " + chr(13)+chr(10) + "WHERE " + cField 
            endif
            first = false
         else
            if this.bIsOr == false  /////////////////////////////////////////////////// Ronnie
               newSQLquery += " " + chr(13)+chr(10) + space(4) + " AND "+cField 
            else
               newSQLquery += " " + chr(13)+chr(10) + space(4) + " OR "+ cField 
            endif
         endif
         if i == this.aConditions.size and (this.bIsNot == true) ///////////////////// Ronnie
            newSQLQuery += " )"
         endif

      next

      // Group by clause
      first = true

      for i = 1 to this.aGroupFields.size
         if "GROUP BY" $ this.aGroupFields[i]
            cString = substr(this.aGroupFields[i],at("(",this.aGroupFields[i])+2)
            cString = substr(cString,1,at(")",cString)-2)

            cString = oStrParse.addQuotes(cString)

            if first = true
               newSQLquery += " " + chr(13)+chr(10) + "GROUP BY " + cString
               first = false
            else
               newSQLquery += ", " + cString
            endif
         endif
      next


      // Order by clause
      first = true
      for i = 1 to this.aOrderFields.size


         if right(trim(this.aOrderFields[i]),4) == "DESC"
            cText = substr(this.aOrderFields[i],1,at(" DESC",this.aOrderFields[i])-1)
            cField = oStrParse.addQuotes(cText) + " DESC"
         else
            cField = oStrParse.addQuotes(this.aOrderFields[i])
         endif
         if first = true
            newSQLquery += " " + chr(13)+chr(10) + "ORDER BY " + cField
            first = false
         else
            newSQLquery += ", " + cField
         endif
      next

      this.SQLvalue = newSQLquery

   return true

   // -------------------------------------

   function SaveQuery(cFileName)

      // Get file name
      if empty(argvector(1))
         if len(this.SQLFileName) == 0
            cFileName = putFile("Save SQL","*.sql","",true)
            if len(cFileName) > 0
               this.SQLFilePath := substr(cFileName, 1, rat("\", cFileName, 1) - 1)
               this.SQLFileName := substr(cFileName, rat("\", cFileName, 1) + 1)
            else
               return false
            endif
         else
            if len(this.SQLFilePath) == 0
               cFileName = this.SQLFileName
            else
               cFileName = this.SQLFilePath + "\" + this.SQLFileName
            endif
         endif
      endif

      // Check for custom edit
      if len(this.CustomSQLValue) > 0 and this.CustomSQLValue <> this.SQLValue
         cSQLText = this.CustomSQLValue
      else
         cSQLText = this.SQLvalue
      endif

      // Check for disallowed terms
      if at("DELETE", upper(cSQLText)) > 0 OR ;
         at("INSERT", upper(cSQLText)) > 0 OR ;
         at("UPDATE", upper(cSQLText)) > 0 OR ;
         at("CREATE TABLE", upper(cSQLText)) > 0 OR ;
         at("ALTER TABLE", upper(cSQLText)) > 0 OR ;
         at("DROP TABLE", upper(cSQLText)) > 0 OR ;
         at("CREATE INDEX", upper(cSQLText)) > 0 OR ;
         at("DROP INDEX", upper(cSQLText)) > 0    
         msgbox("The SQL statements modifying the tables and/or their data" + chr(13) + ;
                "cannot be used in the main query.","SQL query problem",48)
         return false
      endif

      // Check for existing file
      f = new file()
      if f.exists(cFileName) == true
         f.delete(cFileName)
      endif

      // Create new file
      f.create(cFileName)
      f.puts(cSQLText)
      f.close()

      // Update saved flag
      this.SQLisSaved = true

      // Confirm success
      if this.Silent == false
         cText = cFileName + chr(13) + chr(13) ;
               + "has been created"
         msgbox(cText,"Quick Query - Create SQL File")
      endif

   return cFileName

   // -------------------------------------

   function PreviewSQL()

      // Check we have something to preview
      if this.aSelectedFields.size < 1
         msgbox("At least one field must be selected before you can Preview a Query","QQ Preview Error",16)
         return false
      endif

      // Update the Query string
      this.Generate_Query()

      // Get SQL String
      if len(this.CustomSQLValue) > 0
         cString = this.CustomSQLValue   
      else
         cString = this.SQLvalue
      endif

      // Check for disallowed terms
      if at("DELETE", upper(cString)) > 0 OR ;
         at("INSERT", upper(cString)) > 0 OR ;
         at("UPDATE", upper(cString)) > 0 OR ;
         at("CREATE TABLE", upper(cString)) > 0 OR ;
         at("ALTER TABLE", upper(cString)) > 0 OR ;
         at("DROP TABLE", upper(cString)) > 0 OR ;
         at("CREATE INDEX", upper(cString)) > 0 OR ;
         at("DROP INDEX", upper(cString)) > 0
            msgbox("The SQL statements modifying the tables and/or their data" + chr(13) + "cannot be used in the main query.","SQL query problem",48)
            return false
      endif

      // Preview SQL Data using QR3
      cTempTable = getenv("TEMP") + "\QQTemp.dbf"
      this.MakeTable(cTempTable)
      oQR3 = new QR3(form)
      oQR3.Initialise()
      oQR3.SetDataPath(getenv("TEMP"))
      oQR3.SetTable("QQTemp.dbf")
      td = new TableDef()
      td.tableName = cTempTable
      td.load()
      aFields = new array()
      for nCount = 1 to td.fields.size
         aFields.add(td.fields[nCount].fieldName)
      endfor // next
      release object td
      oQR3.SetChosenFields(aFields)
      oQR3.ReportTitle = "Quick Query Data List"
      oQR3.UITitle = "Quick Query Data List"
      oQR3.PreviewReport()
      release object oQR3

   return true

   // -------------------------------------

   function PrintSQL()

      // Check we have something to print
      if this.aSelectedFields.size < 1
         msgbox("At least one field must be selected before you can Preview a Query","QQ Preview Error",16)
         return false
      endif

      // Update the Query string
      this.Generate_Query()

      // Get SQL String
      if len(this.CustomSQLValue) > 0
         cString = this.CustomSQLValue   
      else
         cString = this.SQLvalue
      endif

      // Check for disallowed terms
      if at("DELETE", upper(cString)) > 0 OR ;
         at("INSERT", upper(cString)) > 0 OR ;
         at("UPDATE", upper(cString)) > 0 OR ;
         at("CREATE TABLE", upper(cString)) > 0 OR ;
         at("ALTER TABLE", upper(cString)) > 0 OR ;
         at("DROP TABLE", upper(cString)) > 0 OR ;
         at("CREATE INDEX", upper(cString)) > 0 OR ;
         at("DROP INDEX", upper(cString)) > 0
            msgbox("The SQL statements modifying the tables and/or their data" + chr(13) + "cannot be used in the main query.","SQL query problem",48)
            return false
      endif

      // Print SQL Data using QR3
      cTempTable = getenv("TEMP") + "\QQTemp.dbf"
      this.MakeTable(cTempTable)
      oQR3 = new QR3(form)
      oQR3.Initialise()
      oQR3.SetDataPath(getenv("TEMP"))
      oQR3.SetTable("QQTemp.dbf")
      td = new TableDef()
      td.tableName = cTempTable
      td.load()
      aFields = new array()
      for nCount = 1 to td.fields.size
         aFields.add(td.fields[nCount].fieldName)
      endfor // next
      release object td
      oQR3.SetChosenFields(aFields)
      oQR3.ReportTitle = "Quick Query Data List"
      oQR3.PrintReport()
      release object oQR3

   return true

   // -------------------------------------

   function MakeTable(cFileName)

      // Check for Output File Name
      if empty(cFileName)
         if len(this.OPTableName) > 0
            cFileName = this.OPTableName
         else
            cFileName = putFile("Save Table", "*.dbf", "", true) 
            if len(cFileName) == 0
               return false
            endif
         endif
      endif

      // Update the Query string
      this.Generate_Query()

      // Get SQL String
      if len(this.CustomSQLValue) > 0 and this.CustomSQLValue <> this.SQLValue
         cString = this.CustomSQLValue
      else
         cString = this.SQLvalue
      endif

      // Check for disallowed terms
      if at("DELETE", upper(cString)) > 0 OR ;
         at("INSERT", upper(cString)) > 0 OR ;
         at("UPDATE", upper(cString)) > 0 OR ;
         at("CREATE TABLE", upper(cString)) > 0 OR ;
         at("ALTER TABLE", upper(cString)) > 0 OR ;
         at("DROP TABLE", upper(cString)) > 0 OR ;
         at("CREATE INDEX", upper(cString)) > 0 OR ;
         at("DROP INDEX", upper(cString)) > 0
            msgbox("The SQL statements modifying the tables and/or their data" + chr(13) + "cannot be used in the main query.","SQL query problem",48)
            return false
      endif

      // Set up query
      local q
      q = new QUERY()
      if left(cString,11) == '--Database:'
         // Activate the database
         db = new database()
         db.databaseName := trim(substr(cString,12,at(chr(13),cString)-12))
         db.active := true
         q.database := db
         cString = substr(cString,at("SELECT",cString)) // line 2
      endif
      q.sql = cString  
      q.requestLive = false
      q.active = true

      // Make the table
      u = new UpdateSet()
      u.source = q.rowset
      u.destination = cFileName
      u.copy()

      // Close query
      q.active = false
      q = null
      if this.bUseAlias
         db.active = false
      endif

      // Update QQ
      if not "TEMP" $ upper(cFileName)
         this.OPTableName := cFileName
      endif

      // Confirm success
      if this.Silent == false and (not "TEMP" $ upper(cFileName))
         cText = cFileName + chr(13) + chr(13) ;
               + "has been created"
         msgbox(cText,"Quick Query - Create Table")
      endif

   return true

   // -------------------------------------

endclass

//--------------------------------------------------------------------
//--------------------------------------------------------------------
//--------------------------------------------------------------------

Class StringParser


   // -------------------------------------
   //
   function removeSpace(cStr)
      local string
      string = cStr
      do while at("  ",string)>0
         string = stuff(string,at("  ",string),2, " ")
      enddo
      return string

   // -------------------------------------
   //
   function MaskString(cStr)
   // by Michael Nuwer
      local s
      s = cStr
      nLen = len(s)
      MaskString = ""
      n=1 // starting possition
      bOpen = false
      do while n <= nLen
         chr = substr(s,n,1)

         if not bOpen and (chr==["] or chr == ['])
            // begin of a quoted string
            bOpen = true

         elseif bOpen and (chr==["] or chr == ['])
            // End of a quoted string
            bOpen = false

         elseif bOpen 
            
            chr = "#"

         endif   
         n++
         MaskString += chr 
      enddo
      return(maskString)

   // -------------------------------------
   //
   function addQuotes(cStr)
      // table.field name
      local cField, cTable
      cTable = substr(cStr,1,at(".",cStr)-1)
      cField = substr(cStr,at(".",cStr)+1)
      if at(" ",cField) > 0  or at("/",cField) > 0 ;
         or at("-",cField) > 0  or at("%",cField) > 0


         if at(")",cField) > 0
            cField = ["]+trim(substr(cField,1,at(")",cField)-1)) +[" )]
         else
            cField = ["]+cField+["]
         endif

      elseif lower(cField) $ "order/group/select/left/right/inner/outer/join/from/where/sum/count/min/max/avg/like/starting/containing/between/password/date/month/year/length"
         cField = ["]+cField+["]
      elseif at(")",cField) > 0 and ;
         lower(substr(cField,1,at(")",cField)-1)) $ "order/group/select/left/right/inner/outer/join/from/where/sum/count/min/max/avg/like/starting/containing/between/password/date/month/year/length"

         cField = ["]+trim(substr(cField,1,at(")",cField)-1)) +[" )]

      endif
      return cTable+"."+cField


   // -------------------------------------
   //
   //   By Dan Howard
   function chrSwap(cStr, cSearch, cRepl)
      local c1, c2, n

      c1 = ""
      c2 = cStr
      n = at(cSearch,c2)
      do while n > 0
         c1 += left(c2,n-1) + cRepl 
         c2 = substr(c2,n+len(cSearch))
         n = at(cSearch, c2)
      enddo
      c1 += c2
   return ( c1 )

   // -------------------------------------
   //
   function NumToken(s, cToken)
      // DanH 06/09/2002
      // Modified to be smarter with spaces.
      // only works with a single char token.
      // nTags = this.string.NumToken(c1,"<")
      // Not very smart if the token a space. .i.e. won't handle "Hello    WORLD!"
      local cStr
      cStr = s

      if cToken = " "
         cStr = this.StrTran(cStr, "  ", " ")         
      endif

      local nCount
      nCount = 1
      if at(cToken, cStr) > 0
         do while at(cToken, cStr, nCount) > 0
            nCount++
         enddo
*         nCount --  // Why decrement this? 06/09/2002
//      else
//         return 0
      endif
   return ( nCount-1 )

   // -------------------------------------
   //
   Function atToken(cStr,cToken,nOccurance)
      // returns the postion of the character *after* the token
      local n
      n = at(cToken,cStr,nOccurance)
      if n > 0
         n = n + len(cToken)
      endif
      return n

   // -------------------------------------
   //
   Function atToken2(cStr,cToken,nOccurance)
      // returns the postion of the character *after* the token
      // taking into account multiple spaces.

      if nOccurance =0
         return 1
      endif

      local n, nn
      nn = nOccurance
      n = 0
      local j,nToken
      nToken = 1
      j = 1
      do while true
         n = at(cToken,cStr,j)
         if n < len(cStr) and n > 0
            if substr(cStr,n+1,1) # cToken
               nToken++
               if nToken = nOccurance
                  n = n+1
                  exit
               endif
            endif
         else
            n = 0
            exit
         endif
         j++
      enddo
   return n

   // -------------------------------------
   //
   // Dirk lesko's StrTran..............
   FUNCTION StrTran
      PARAM cTxt,cOrgTxt,cRepTxt,nStart,nCount
      LOCAL nCx,nIx,cNewStr, nOldStrLen, nStrLen

      IF ArgCount() < 3
         cRepTxt = ''
      ENDIF

      IF ArgCount() < 4
         nStart=1
      ENDIF
      IF(nStart < 1)
       return cTxt
      ENDIF

      nStrLen = Len(cTxt)
      IF ArgCount()<5
         nCount=nStrLen
      ENDIF
      IF( nCount < 0 )
       return cTxt
      ENDIF

      IF Type('cTxt')#'C'
       IF Type('cTxt')#'M'
         return cTxt // let them blow up if they use it...
         ENDI
      ENDIF

      IF Type('cRepTxt')#'C'
        cRepTxt=''
      ENDIF

      cNewStr=''
      nOldStrLen = Len(cOrgTxt)

      nIx=At(cOrgTxt,cTxt,nStart)
      DO WHILE( nIx > 0 .And. nCount-- > 0)

         cNewStr=cNewStr + Substr(cTxt,1,nIx-1) + cRepTxt
         cTxt = Substr(cTxt, nIx + nOldStrLen, nStrLen)
         nIx=At(cOrgTxt,cTxt, 1)

      ENDDO
   RETURN cNewStr + cTxt

   // -------------------------------------
   //
   // Function:      StrExtract
   // Purpose:       Extracts a substring from a string with
   //                a specified delimiter.
   // Parameters:
   //       cStr   - string to search
   //       cToken - item delimiter
   //       nItem  - item number to extract
   function StrExtract(cStr, cToken, nItem)
   local p1,p2,j
   local cItem
   local cSubStr
      private c
      c = cStr
      if type("c") <> "C"
         return ""
      endif

      p1 = 1
      p2 = 0
      cSubStr = cStr
      cItem  = ""
      j = 1
      do while .not. empty(cSubStr)
         pos = at(cToken,cSubStr)-1
         if pos <= 0
            pos = len(cSubStr)
         endif
         cItem = left(cSubstr,pos)
         if j = nItem
            exit
         endif
         if pos < len(cSubStr) - 1
            cSubStr = substr(cSubStr,pos+2,len(cSubStr))
            if cToken == " "
               cSubStr = ltrim(cSubStr)
            endif
         else
            cSubStr = ""
            cItem   = ""
         endif
         j = j + 1
      enddo
   return (cItem)


endclass

// -------------------------------------------------------------------
// -------------------------------------------------------------------
// -------------------------------------------------------------------

//EOF