/* dCGISession A class that mangages CGI operations when dBASE is used  to
					create Web Applets. The current version supports only CGI.
               Future versions will support ISAPI and NSAPI and a likely
               dBASE application server.

   Usage:		oCGI = new CGISession()

	Author:		A.A.Katz, author
               v.1.0 11/27/1999
               v.1.1 01/12/2000 (supports win 98 StdOut)
   Rights:		You have the right to freely include this code in any compiled
   				dBASE Web application. You do not have the right to reproduce
               or publish this source code without the explicit written 
               permission of dBASE, Inc., Vestal, NY
               (c) 1999-2000 dBASE Inc.
   Platform:	This code works only with Visual dBASE 7.5 or later.
   				If this code is used, you do not need to use VdB_CGI.exe
               as required with earlier versions. Your Visual dBASE
               applications may be called directly by the Web Server.
*/


Class CGISession of assocArray
   ////// Constructor code


   ////// Set dBASE environment commands. Web apps are non-GUI

	SET CENTURY ON
   SET TALK OFF


   
  
   ////// Method:		Connect //////////////////////////////////////////////////
   ////// Purpose		Connects by StdIn and StdOut to Web server ///////////////
   //////            reads in data, and then parses out name/value pairs //////
   //////				and creates a new element of this array for each. ////////

   Function Connect
   
      // Read incoming data into this array as name/value pairs
      // Ex: this["City] = "Vestal", where "City" is the name
      // and "Vestal" is the value
  
      this.loadArrayFromCGI()  

      ////// Open  output stream back to Web Server
      Try

         this.fOut = new file()
         this.fOut.Open("StdOut", "RA")  // open the StdOut output stream
         
      Catch (exception e)

         Quit 							// Might as well quit, can't send response

      endTry


   ////// Method:  LoadArrayFromGGI //////////////////////////////////////
   ////// Purpose: Loads existing assoc array with data from  ////////////
   //////  			 Web Server from a previous page////////////////////////
   ////// Params:  None //////////////////////////////////////////////////
 				
   /* Note: This code connects to the Web Server and captures the incoming
   			data sent when the HTML form was submitted. This Associative
            Array, fortunately, mimics the Web's Name/Value pairs.
            Example: this["City"] = "Vestal", where "City" is the name and
           "Vestal" is the value. 
   */

   Function LoadArrayFromCGI

      // First, look in environment to see if the page was submitted via 
      // Post or Get

      cMethod = upper(getEnv("REQUEST_METHOD"))

      if cMethod = 'POST'    // if the submit method was "Post" (Recommended!!!) 
                             // get length of incoming data string
         nLen = val(getEnv("CONTENT_LENGTH")) 

         Try

            fIn  = new file()        // Create new instance of File Object
            fIn.Open("StdIn", "RA")  // open the StdIn input stream
            cEnv = fIn.Read(nLen)    // read content into "cEnv"  variable

         Catch (exception e)

            Quit        				// Might as well quit, can't open connection

         endTry

         // Note: StdIn and StdOut cannot be close()'ed.
          
      else

         cEnv = getEnv("QUERY_STRING") // Get data from environment

      endif

      /* Note: Name/Value pairs are sent in a form something like this:
             	City=Vestal&State=NY&Country=USA  with "&" as delimeter
               between named pairs.

         Note: Punctuation and spaces are sent using special characters.
         		Furthermore, the Internet uses the International character
               standard: ANSI while dBASE uses its own internal character
               standard: OEM. The OEMFormat() method handles all
               conversions to dBASE-usable character strings automatically.
      */

      ////// Break up CGI input string into 
      ////// Name/Value pairs

      try      

         do while at("=",cEnv) > 0   // Do while there are still pairs

            cName  = ''             // init name/value vars
            cValue = ''

                      
            If at("&",cEnv) > 0      // If there is more than one pair 
                                     // Grab the pair before first "&'
               cEnvPair = this.OEMFormat(substr(cEnv,1,at("&",cEnv)-1))
                  
                                     // Strip this pair off of cEnv
                                     // in preparation for next loop
               cEnv = substr(cEnv,at("&",cEnv)+1)

            else                     // This must be the last pair

               cEnvPair = this.OEMFormat(cEnv) // grab remainder
               cEnv = ''             // Nothing left to parse

            endif           

            ////// separate name and value
            cName = substr(cEnvPair,1,at('=',cEnvPair)-1)+""
            cValue = substr(cEnvPair, at('=',cEnvPair)+1)+""

            ////// if name/value pair exists, append into this array
            if len(trim(cName)) >0 // 

               this[cName] = cValue  // update or add to array

            endif
            
         enddo

      catch (exception e)

         this.errorPage(e)

      endtry
        
      return true


   ////// Method 		OEMFormat /////////////////////////////////////////////////
   ////// Purpose 	Transforms CGI escape chars to OEM/////////////////////////
   ////// Param: 		cStr 	Character strings sent by Web browser ///////////////
   //////			   in ANSI format ////////////////////////////////////////////

   /* Note: Strings come in over the internet in ANSI format and must be
            converted to dBASE OEM Format before use. Punctuation is 
            sent in %nn hexadecimal format and spaces to plus signs (+).
            This method does all the conversion necessary to convert from
            ANSI CGI to dBASE OEM.
            Example: Alan's New Class, in ANSI CGI is 
                     Alan%27s+New+Class
   */

   Function OEMFormat(cStr) Protect // For internal use only

      // remove "Pluses" used as space keepers in CGI
      do while at("+",cStr) > 0
         cStr = substr(cStr,1,at("+",cStr)-1) + " "+;
                substr(cStr,at("+",cStr)+1)
      enddo

      // add in "CRs" sent as escape characters
      do while at("%0D%0A",cStr) > 0
         cStr = substr(cStr,1,at("%0D%0A",cStr)-1) +chr(13)+;
                substr(cStr,at("%0D%0A",cStr)+3)
      enddo

      // convert escape characters back to oem chars
      nLeft = 1
      do while at("%",substr(cStr,nLeft) ) > 0
         nPos = at("%",substr(cStr,nLeft))+nLeft-1
         nLeft = nPos+1
         if nPos <= len(cStr)-2
            cStr = left(cStr,nPos-1)+ chr(htoi(substr(cStr,nPos+1,2)))+;
            substr(cStr,nPos+3)
         endif
      enddo
 
      return cStr


  


   /////////////////////// Output Methods //////////////////////////////////



   ////// Method:   SorryPage //////////////////////////////////////////////
   ////// Purpose:    Returns Generic Sorry response page //////////////////
   ////// Param:      cMsg  = Error Message ////////////////////////////////
   //////             		   Accepts a char string or array of messages ///
   //////             cSubTtl Subtitle, such as 'An error has ocurred"  or /
   //////                    "data missing" ////////////////////////////////
   //////             cRecover = Suggestion for recovery ///////////////////
   //////                     such as "Press the browser's back button" ////
   //////                     or contact Web master.... ////////////////////

   Function SorryPage(cMsg,cSubTtl,cRcvr)
  
      aMsg = iif(empty(cMsg),'',cMsg)

      if type('aMsg') = 'C'   // if message is not array,
         aMsg = new array()   // convert to a one-element array
         aMsg.add(cMsg)
      endif
                               // Make private versions for type()
      cRecover =  iif(empty(cRcvr),'Press Back button and try again.',cRcvr)  
      cSubtitle = iif(empty(cSubTtl),'',cSubTtl)


      /////// Stream out header
      this.StreamHeader('Sorry!')

      with (this.fOut)

         ////// Body tag (starts body of page)  
         puts('<BODY>')
         puts('')
         puts('<P>')
         puts('<TABLE BORDER="0" CELLPADDING="0" CELLSPACING="0" WIDTH="100%">')

         ////// Sorry!
         puts('	<TR>')
         puts('		<TD COLSPAN="2" BGCOLOR="#E4E4E4">')
         puts('			<H1><FONT COLOR="#999999">Sorry!</FONT>')
         puts('		</TD>')
         puts('	</TR>')

         ////// Blank Line
         puts('	<TR>')
         puts('		<TD WIDTH="9%">&nbsp;</TD>')
         puts('		<TD WIDTH="91%">&nbsp;</TD>')
         puts('	</TR>')

      ////// Subtitle

         if len(trim(cSubTitle)) > 0 // if there's a subtitle
    
            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%"><I><FONT SIZE="3" FACE="Arial, Helvetica">'+cSubTitle+'</FONT></I></TD>')
            puts('	</TR>')
            puts('	<TR>')

            ////// Blank Line
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%">&nbsp;</TD>')
            puts('	</TR>')

  
         endif

      ////// Traverse the message array
      	for n = 1 to aLen(aMsg,1)

            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%"><FONT SIZE="2" FACE="Arial, Helvetica">'+aMsg[n]+'</FONT></TD>')
            puts('	</TR>')

         next

         ////// If recovery instructions exist (Contact Webmaster, Press Back
         ////// 		Button, etc.), stream it here
         if len(trim(cRecover)) > 0
            with (this.fOut)
            ////// Blank line
            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%">&nbsp;</TD>')
            puts('	</TR>')

            ////// Recovery message
            puts('	<TR>')
            puts('		<TD WIDTH="9%">&nbsp;</TD>')
            puts('		<TD WIDTH="91%"><FONT SIZE="2" FACE="Arial, Helvetica">'+cRecover+'</FONT></TD>')
            puts('	</TR>')
            endwith
         endif

   
         puts('</TABLE>')
         puts('')
         puts('')
         puts('</BODY>')


      endwith

      ////// Close out HTML structure
      this.StreamFooter()
 
      quit


   /////// Method:	errorPage  ////////////////////////////////////////////////
   /////// Purpose:	Reports error back to user and quits application //////////
   /////// Params:	oErr - takes the error object thrown by the dBASE /////////
   ///////				error-handling system. Can be thrown by the developer, ////
   ///////				or passed from the "Catch" of a try...catch statement /////
   /////// Calls:		this.sorryPage to stream error page back to Server ////////

   Function errorPage(oErr)

      aError = new array()                // Create array for messages

      with (aError)
         add('File: '+oErr.Filename)     // Add file/line and message
         add('Line No:'+oErr.LineNo)
         add(''+oErr.Message)
      endwith

 
      cRecover = 'Contact the WebMaster'
     
                                          // Call sorryPage() to display  
      this.sorryPage(aError,'An error ocurred on the server',cRecover)

      return true





  ////////////////////////// Database Classes //////////////////////////////////


   ////// Method: loadDatamoduleFromArray ////////////////////////////
   //////         loads data in this array into the rowset.fields ////
   //////         of the datamodule
   ////// Param:	oDataMod - a reference to a DataModule object ///////
   ////// 			bAppend - create new row?

   Function loadDataModuleFromArray(oDataMod,bAppend)
  
      if empty(oDataMod)
         return false // no datamodule to work with
      endif

      oD = oDataMod  // "Privatize param"  for use in macros

                     // ensure default of bAppand
      bAppend = iif(empty(bAppend),true,bAppend)

      ////// This array will track which queries have already
      ////// been appended so we don't create extra rows  or
      ////// save rows unneccessarily

      aAppends = new array()

      ////// Traverse this array of field/value pairs and update
      ////// rowset fields with data

      cKey = this.firstKey  // init first key 

         
      For n = 1 to this.Count()  // traverse entire array

         if at('*@',cKey) = 0    // if not a datamod pair
            cKey = this.nextKey(cKey)
            loop                 // do next one
         endif
  
         ////// extract query name
         cQuery = substr(cKey,1,at("*@",cKey)-1)

     
         ////// Prepare macro to test query
         cStr = 'oD.'+cQuery
           
         ////// See if query exists
         Try
            oQuery = &cStr.
         catch (exception e)
            // this query does not exist!
         
            cKey = this.nextKey(cKey)
            loop
         endTry


         if type('oQuery.rowset') # 'O'
            // Rowset not valid???
            cKey = this.nextKey(cKey)
            loop
         endif

         ////// See if this query has been appended or edited
         ////// if not, append row if bAppend = true
         if aAppends.scan(upper(cQuery))  = 0

            aAppends.add(upper(cQuery)) // add to query array
            if bAppend      
               oQuery.rowset.beginAppend()
            endif

         endif

         ////// extract field name
         cField = trim(substr(cKey,at("*@",cKey)+2))

         if type('oQuery.rowset.fields[cField]') # 'O' // if no such field exists

            // loop and try the next
            cKey = this.nextKey(cKey)
            loop
         endif


         with (oQuery.rowset.fields[cField])       // with (field) 

            ////// Convert from string to type and update field value    
       
            if type 	= "LOGICAL"

               value  = ;
                  iif(upper(this[cKey]) = 'TRUE' or upper(this[cKey]) = 'Y',;
                  true,false)

            elseif type = "NUMERIC" or ;
               type = "INTEGER" or ;
               type = "LONG" or ;
               type = "DOUBLE" or ;
               type = "FLOAT"

               value = val(this[cKey])

            elseif type = "DATE"

               value = ctod(this[cKey])

            elseif type = 'AUTOINCREMENT' or ;
                   type = 'TIMESTAMP'  or ;
                   type = 'BINARY' or ;
                   type = 'OLE'

               	// do nothing, not supported in HTML


            else 	// char or memo   

               value = this[cKey]
                 
            endif

         endwith

         cKey = this.nextKey(cKey)
             
      Next

      ////// Traverse all queries and save rows
      for n = 1 to alen(aAppends,1) 
         cQuery = aAppends[n]

         cStr = 'oD.'+cQuery  // Prepare string for macro

         try
            oQuery = &cStr.   // Get reference for this query
         catch (exception e)
            loop              // if fail, go to next
         endTry


         if not oQuery.rowset.endOfSet
            oQuery.rowset.modified = true
            oQuery.rowset.save() // save row.
         endif

      next


  ////// Method: StreamFooter /////////////////////////////////////////////////
  //////         Streams out closing tags for header/title block for //////////
  //////         HTML Response page sent back to user//////////////////////////
 
   Function streamFooter

      with (this.fOut)
         this.fOut.puts(' ')
         this.fOut.puts('   </BODY>')
         this.fOut.puts(' ')
         this.fOut.puts('</HTML>')
      endwith
      return true


 

   function streamResponsePage(cBack,cBar,cLogo)

      with (this.fOut)
         puts('Content-type: text/html')
         puts('')
 
         puts('<html>')
         puts('<head>')
         puts('<title>dQuery/Web</title>')
         puts('<meta http-equiv="Content-Type" content="text/html; charset=iso-8859-1">')
         puts('</head>')
         puts('<body bgcolor="'+cBack+'" link="#000000" vlink="#000000" alink="#000000">')
         puts('<table width="74%" border="0" cellpadding="2">')
         puts('  <tr> ')
         puts('    <td rowspan="2"><img src="'+cLogo+'"></td>')
         puts('    <td width="92%">')
         puts('      <table width="100%" border="0" cellpadding="2">')
         puts('        <tr> ')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td bgcolor="'+cBar+'">&nbsp;</td>')
         puts('          <td>&nbsp;</td>')
         puts('        </tr>')
         puts('      </table>')
         puts('    </td>')
         puts('  </tr>')
         puts('  <tr> ')
         puts('    <td width="92%">&nbsp;</td>')
         puts('  </tr>')
         puts('</table>')
         puts('<table width="74%" border="0" cellpadding="2" height="116">')
         puts('  <tr> ')
         puts('    <td width="4%" height="52">&nbsp;</td>')
         puts('    <td width="4%" bgcolor="'+cBar+'" height="52">&nbsp;</td>')
         puts('    <td width="92%" align="center" height="52"> ')
         puts('      <blockquote>')
         puts('        <h2 align="left"><font color="#999999">Data Entered Successfully!</font></h2>')
         puts('      </blockquote>')
         puts('    </td>')
         puts('  </tr>')
         puts('  <tr> ')
         puts('    <td width="4%" height="46">&nbsp;</td>')
         puts('    <td width="4%" bgcolor="'+cBar+'" height="46">&nbsp;</td>')
         puts('    <td width="92%" align="center" height="46"> ')
         puts('      <blockquote> ')
         puts([        <p align="left"><font face="Arial, Helvetica, sans-serif">Press the Browser's ])
         puts('          Back button to<br>')
         puts('          add another entry. </font><font face="Arial, Helvetica, sans-serif">Click ')
         puts('          &quot;Reset&quot; to clear out the form...</font></p>')
         puts('      </blockquote>')
         puts('    </td>')
         puts('  </tr>')
         puts('  <tr> ')
         puts('    <td width="4%" height="25">&nbsp;</td>')
         puts('    <td width="4%" bgcolor="'+cBar+'" height="25">&nbsp;</td>')
         puts('    <td width="92%" align="center" height="25">&nbsp;</td>')
         puts('  </tr>')
         puts('</table>')
         puts('</body>')
         puts(' ')
         puts('   </BODY>')
         puts(' ')
         puts('</HTML>')

	   endwith

      this.fOut.close()

      return true


   ////// Method: 	StreamHeader //////////////////////////////////////////////
   ////// Purpose		Streams the CGI and HTML headers required  ////////////////
   //////				to send a response page back through the server ///////////
   ////// Param:		Title - the page title desired ////////////////////////////

   Function StreamHeader(Title)
      cTitle = iif(empty(Title),'Response Page',Title)

      with (this.fOut)
         puts('Content-type: text/html')
         puts('')
         puts('<HTML>')
         puts('')
         puts('<HEAD>')
         puts('	<META HTTP-EQUIV="Content-Type" CONTENT="text/html">')
         puts('	<TITLE> '+cTitle+'</TITLE>')
         puts('</HEAD>')
         puts('')
      endwith

      return true
 

   ////// Method:    CheckRequiredFields ///////////////////////////////////////
   ////// Purpose:	Validates existence of required data as
   /////					defined by the user/developer
   ////// Params:		aRequired an array listing required fields

   Function checkRequiredFields(aRequired)
      private n, cKey
    
      // traverse the "required" array
      for n = 1 to aRequired.size

         // Get the key (query*@fieldname)
         cKey = aRequired[n]

         // See if it's a field on this data-entry form
         if this.isKey(cKey)

            // If it is, see if it's empty
            if empty(this[cKey])

               // If it is, give error and quit
               cField = substr(cKey,at('*@',cKey)+2)
               this.sorryPage('Field:'+cField+' is required!',;
                    'Missing Data')
            endif

         endif

      next

      return true

 EndClass



